#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import time
import mock

from tests import event_loop
import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums

from octobot_trading.api.exchange import cancel_ccxt_throttle_task

from tests.exchanges import simulated_exchange_manager, simulated_trader
from tests.personal_data.trades import create_executed_trade


SYMBOL = "BTC/USDT"


def test_aggregate_trades_by_exchange_order_id(simulated_trader):
    _, _, trader = simulated_trader
    t1 = create_executed_trade(
        trader, enums.TradeOrderSide.BUY, 1, decimal.Decimal(10), decimal.Decimal(1500), SYMBOL, _get_fees("BTC", 0.1)
    )
    t2 = create_executed_trade(
        trader, enums.TradeOrderSide.BUY, 3, decimal.Decimal(13), decimal.Decimal(1605), SYMBOL, _get_fees("BTC", 0.5)
    )
    t3 = create_executed_trade(
        trader, enums.TradeOrderSide.BUY, 7, decimal.Decimal(7), decimal.Decimal(1400), SYMBOL, _get_fees("BTC", 1.4)
    )
    aggregated = personal_data.aggregate_trades_by_exchange_order_id([t1, t2, t3])
    # aggregated all trades together as all of their exchange_order_id is None
    assert list(aggregated) == [None]
    assert aggregated[None].trader is trader
    assert aggregated[None].symbol == t1.symbol
    assert aggregated[None].side is t1.side
    assert aggregated[None].executed_price == decimal.Decimal("1522.166666666666666666666667")
    assert aggregated[None].executed_quantity == sum(t.executed_quantity for t in (t1, t2, t3))
    assert aggregated[None].total_cost == sum(t.total_cost for t in (t1, t2, t3))
    assert aggregated[None].fee[enums.FeePropertyColumns.COST.value] == \
       sum(t.fee[enums.FeePropertyColumns.COST.value] for t in (t1, t2, t3))
    assert aggregated[None].executed_time == 7

    # set exchange_order_id
    t1.exchange_order_id = "1"
    t2.exchange_order_id = "1"
    t3.exchange_order_id = "2"
    aggregated = personal_data.aggregate_trades_by_exchange_order_id([t1, t2, t3])
    # aggregated all trades together as all of their exchange_order_id is None
    assert list(aggregated) == ["1", "2"]
    assert aggregated["1"].trader is trader
    assert aggregated["1"].symbol == t1.symbol
    assert aggregated["1"].side is t1.side
    assert aggregated["1"].executed_price == decimal.Decimal("1559.347826086956521739130435")
    assert aggregated["1"].executed_quantity == sum(t.executed_quantity for t in (t1, t2))
    assert aggregated["1"].total_cost == sum(t.total_cost for t in (t1, t2))
    assert aggregated["1"].fee[enums.FeePropertyColumns.COST.value] == \
       sum(t.fee[enums.FeePropertyColumns.COST.value] for t in (t1, t2))
    assert aggregated["1"].executed_time == 3

    assert aggregated["2"].executed_price == decimal.Decimal(1400)
    assert aggregated["2"].executed_quantity == decimal.Decimal(7)
    assert aggregated["2"].total_cost == t3.total_cost
    assert aggregated["2"].fee == _get_fees("BTC", 1.4)
    assert aggregated["2"].executed_time == 7


def _get_fees(currency, value):
    return {
        enums.FeePropertyColumns.CURRENCY.value: currency,
        enums.FeePropertyColumns.COST.value: decimal.Decimal(str(value))
    }


class TestGetTradingOrderFee:
    """Test class for get_trading_order_fee method."""

    def create_trade(
        self, trader, exchange_order_id: str,
        side, executed_quantity, executed_price, symbol, fee
    ) -> personal_data.Trade:
        return personal_data.create_trade_from_dict(
            trader,
            {
                enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value: exchange_order_id,
                enums.ExchangeConstantsOrderColumns.SYMBOL.value: symbol,
                enums.ExchangeConstantsOrderColumns.SIDE.value: side,
                enums.ExchangeConstantsOrderColumns.AMOUNT.value: executed_quantity,
                enums.ExchangeConstantsOrderColumns.PRICE.value: executed_price,
                enums.ExchangeConstantsOrderColumns.TYPE.value: enums.TradeOrderType.LIMIT.value,
                enums.ExchangeConstantsOrderColumns.COST.value: executed_quantity * executed_price,
                enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: time.time(),
                enums.ExchangeConstantsOrderColumns.FEE.value: fee,
            }
        )

    def test_gets_fee_from_trade_with_base_currency(self, simulated_trader):
        _, _, trader = simulated_trader
        # Create trade with fee already set (base currency)
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL,
            {
                enums.FeePropertyColumns.CURRENCY.value: "BTC",
                enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),  # 0.1%
            }
        )
        
        # Get trading order fee
        # Should return calculated fee from trade.fee directly and not be estimated
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            _get_fees("BTC", 0.0001),
            False
        )

    def test_gets_fee_from_trade_with_quote_currency(self, simulated_trader):
        _, _, trader = simulated_trader
        # Create trade with fee already set (quote currency)
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL,
            {
                enums.FeePropertyColumns.CURRENCY.value: "USDT",
                enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),  # 0.1%
            }
        )
        
        # Get trading order fee
        # Should return calculated fee from trade.fee directly and not be estimated
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            _get_fees("USDT", 0.1),
            False
        )

    def test_gets_fee_from_order_with_base_currency(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        # Create mock order with fee in base currency
        mock_order = mock.Mock()
        mock_order.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "BTC",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),  # 0.1%
        }
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Get trading order fee
        # Should return calculated fee and not be estimated
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            _get_fees("BTC", 0.0001),
            False
        )

    def test_gets_fee_from_order_with_quote_currency(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        # Create mock order with fee in quote currency
        mock_order = mock.Mock()
        mock_order.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),  # 0.1%
        }
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Get trading order fee
        # Should return calculated fee and not be estimated
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            _get_fees("USDT", 0.1),
            False
        )
            
    def test_gets_fee_from_trades_when_order_has_no_fee(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        
        # Create mock order without fee
        mock_order = mock.Mock()
        mock_order.fee = None
        
        # Create mock trade with fee
        mock_trade = mock.Mock()
        mock_trade.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),
        }
        
        # Mock orders_manager.get_order to return order without fee
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Mock trades_manager.get_trades to return trade with fee
        trader.exchange_manager.exchange_personal_data.trades_manager.get_trades = mock.Mock(
            return_value=[mock_trade]
        )
        
        # Get trading order fee
        # Should return calculated fee from trade and not be estimated
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            # Fee cost = rate * amount * price = 0.001 * 0.1 * 1000 = 0.1 USDT
            _get_fees("USDT", 0.1),
            False
        )

    def test_gets_fee_from_trades_when_order_fee_missing_fields(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        
        # Create mock order with incomplete fee (missing RATE)
        mock_order = mock.Mock()
        mock_order.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            # Missing RATE
        }
        
        # Create mock trade with fee
        mock_trade = mock.Mock()
        mock_trade.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "BTC",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),
        }
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Mock trades_manager.get_trades
        trader.exchange_manager.exchange_personal_data.trades_manager.get_trades = mock.Mock(
            return_value=[mock_trade]
        )
        
        # Get trading order fee
        # Should return calculated fee from trade and not be estimated
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            # Fee cost = rate * amount = 0.001 * 0.1 = 0.0001 BTC
            _get_fees("BTC", 0.0001),
            False
        )

    def test_skips_trades_without_fee(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        
        # Create mock order without fee
        mock_order = mock.Mock()
        mock_order.fee = None
        
        # Create mock trades: one without fee, one with fee
        mock_trade_no_fee = mock.Mock()
        mock_trade_no_fee.fee = None
        
        mock_trade_with_fee = mock.Mock()
        mock_trade_with_fee.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),
        }
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Mock trades_manager.get_trades to return both trades
        trader.exchange_manager.exchange_personal_data.trades_manager.get_trades = mock.Mock(
            return_value=[mock_trade_no_fee, mock_trade_with_fee]
        )
        
        # Get trading order fee
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            # Should return calculated fee from the trade with fee
            _get_fees("USDT", 0.1),
            False
        )

    def test_estimates_fee_when_no_order_or_trade_fee(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        
        # Create mock order without fee
        mock_order = mock.Mock()
        mock_order.fee = None
        
        # Create mock trades without fee
        mock_trade = mock.Mock()
        mock_trade.fee = None
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Mock trades_manager.get_trades
        trader.exchange_manager.exchange_personal_data.trades_manager.get_trades = mock.Mock(
            return_value=[mock_trade]
        )
        
        # Mock get_trade_fee
        mock_fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            enums.FeePropertyColumns.COST.value: decimal.Decimal("0.1"),
        }
        with mock.patch.object(
            trader.exchange_manager.exchange,
            "get_trade_fee",
            return_value=mock_fee
        ) as mock_get_trade_fee:
            # Get trading order fee
            # Should estimate fee and return True for is_estimated
            assert personal_data.get_real_or_estimated_trade_fee(
                trade
            ) == (
                mock_fee,
                True
            )
            
            # Verify get_trade_fee was called with correct parameters
            mock_get_trade_fee.assert_called_once_with(
                trade.symbol,
                enums.TraderOrderType.BUY_LIMIT,  # BUY trade -> BUY_LIMIT
                trade.executed_quantity,
                trade.executed_price,
                trade.taker_or_maker
            )

    def test_handles_order_not_found_exception(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.1"), decimal.Decimal("1000"), SYMBOL, None
        )
        
        # Mock orders_manager.get_order to raise KeyError (order not found)
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            side_effect=KeyError("Order not found")
        )
        
        # Create mock trade with fee
        mock_trade = mock.Mock()
        mock_trade.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),
        }
        
        # Mock trades_manager.get_trades
        trader.exchange_manager.exchange_personal_data.trades_manager.get_trades = mock.Mock(
            return_value=[mock_trade]
        )
        
        # Get trading order fee
        # Should fall back to trade fee
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            _get_fees("USDT", 0.1),
            False
        )

    def test_calculates_base_fee_correctly(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.BUY, decimal.Decimal("0.5"), decimal.Decimal("1000"), SYMBOL, None
        )
        
        # Create mock order with fee in base currency
        mock_order = mock.Mock()
        mock_order.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "BTC",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.002"),  # 0.2%
        }
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Get trading order fee
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            _get_fees("BTC", 0.001),
            False
        )

    def test_calculates_quote_fee_correctly(self, simulated_trader):
        _, _, trader = simulated_trader
        trade = self.create_trade(
            trader, "order_123", enums.TradeOrderSide.SELL, decimal.Decimal("0.2"), decimal.Decimal("5000"), SYMBOL, None
        )
        
        # Create mock order with fee in quote currency
        mock_order = mock.Mock()
        mock_order.fee = {
            enums.FeePropertyColumns.CURRENCY.value: "USDT",
            enums.FeePropertyColumns.RATE.value: decimal.Decimal("0.001"),  # 0.1%
        }
        
        # Mock orders_manager.get_order
        trader.exchange_manager.exchange_personal_data.orders_manager.get_order = mock.Mock(
            return_value=mock_order
        )
        
        # Get trading order fee
        assert personal_data.get_real_or_estimated_trade_fee(
            trade
        ) == (
            # Fee cost = rate * amount * price = 0.001 * 0.2 * 5000 = 1.0 USDT
            _get_fees("USDT", 1.0),
            False
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchanges.connectors import ccxt
from octobot_trading.exchanges.connectors.ccxt import (
    CCXTAdapter,
    CCXTConnector,
    CCXTWebsocketConnector,
)

from octobot_trading.exchanges.connectors import simulator
from octobot_trading.exchanges.connectors.simulator import (
    ExchangeSimulatorAdapter,
    ExchangeSimulatorConnector,
)

from octobot_trading.exchanges.connectors import util
from octobot_trading.exchanges.connectors.util import (
    retried_failed_network_request,
)

__all__ = [
    "CCXTAdapter",
    "CCXTConnector",
    "CCXTWebsocketConnector",
    "ExchangeSimulatorAdapter",
    "ExchangeSimulatorConnector",
    "retried_failed_network_request",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import decimal

import octobot_trading.modes.script_keywords as script_keywords
import octobot_trading.modes.script_keywords.basic_keywords.position as position
import octobot_trading.enums as enums
import octobot_trading.constants as constants

from tests import event_loop
from tests.modes.script_keywords import mock_context
from tests.exchanges import backtesting_trader, \
    backtesting_config, backtesting_exchange_manager, fake_backtesting


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.mark.parametrize("backtesting_config", ["USDT"], indirect=["backtesting_config"])
async def test_average_open_pos_entry(mock_context):
    mock_context.exchange_manager.is_future = False
    with mock.patch.object(position, "get_position",
                           mock.Mock(return_value=(mock.Mock(entry_price=decimal.Decimal("40000"))))) \
         as get_position_mock:
        mock_context.symbol = "ETH/USDT"
        assert await script_keywords.average_open_pos_entry(mock_context, enums.TradeOrderSide.BUY.value) == \
               constants.ZERO
        mock_context.exchange_manager.is_future = True
        mock_context.symbol = "ETH/USDT"
        assert await script_keywords.average_open_pos_entry(mock_context, enums.TradeOrderSide.BUY.value) == \
               decimal.Decimal("40000")
        get_position_mock.assert_called_once_with(mock_context, "ETH/USDT", enums.TradeOrderSide.BUY.value)
        get_position_mock.reset_mock()
        mock_context.symbol = "BTC/USDT"
        assert await script_keywords.average_open_pos_entry(mock_context, enums.TradeOrderSide.BUY.value) == \
               decimal.Decimal("40000")
        get_position_mock.assert_called_once_with(mock_context, "BTC/USDT", enums.TradeOrderSide.BUY.value)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import octobot_trading.personal_data.orders.cancel_policies.order_cancel_policy as order_cancel_policy_import


@dataclasses.dataclass
class ExpirationTimeOrderCancelPolicy(order_cancel_policy_import.OrderCancelPolicy):
    """
    Will cancel the order if the expiration time is reached.
    """
    expiration_time: float

    def should_cancel(self, order) -> bool:
        if order.is_cleared():
            self.get_logger().error(
                f"Ignored cancel policy: order {str(order)} has been cleared"
            )
            return False
        return self.expiration_time <= order.trader.exchange_manager.exchange.get_exchange_current_time()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest
import asyncio

import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.order_factory as order_factory
import octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group as \
    balanced_take_profit_and_stop_order_group
import octobot_trading.personal_data.orders.groups.trailing_on_filled_tp_balanced_order_group as \
    trailing_on_filled_tp_balanced_order_group
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_util as order_util
from tests.exchanges import backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests.personal_data.orders.groups import order_mock
from tests.personal_data import DEFAULT_ORDER_SYMBOL
from tests.personal_data.orders import created_order
from tests.exchanges import simulated_trader, simulated_exchange_manager


@pytest.fixture
def toftpb_group(backtesting_exchange_manager):
    orders_manager = mock.Mock()
    orders_manager.get_order_from_group = mock.Mock()
    return personal_data.TrailingOnFilledTPBalancedOrderGroup("name",  orders_manager)


@pytest.fixture
def side_balance():
    return balanced_take_profit_and_stop_order_group.SideBalance(None, False)


@pytest.fixture
def trailing_side_balance():
    return trailing_on_filled_tp_balanced_order_group.TrailingSideBalance(None, True)


@pytest.fixture
def order_f():
    trader = mock.Mock(exchange_manager=mock.Mock())
    target_order = personal_data.Order(trader)
    return target_order


def test_balances_factory(toftpb_group):
    closed_orders = []
    filled = False
    balances = toftpb_group._balances_factory(closed_orders, filled)
    assert isinstance(balances[toftpb_group.TAKE_PROFIT], balanced_take_profit_and_stop_order_group.SideBalance)
    assert isinstance(balances[toftpb_group.STOP], trailing_on_filled_tp_balanced_order_group.TrailingSideBalance)


@pytest.mark.asyncio
async def test_trailing_balanced_trail_after_order_fill(simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    trader_instance.allow_artificial_orders = False
    trader_instance.enable_inactive_orders = True
    trader_instance.simulate = False
    exchange_manager.is_backtesting = True  # force update_order_status
    tbtp_group = personal_data.TrailingOnFilledTPBalancedOrderGroup(
        "name", exchange_manager.exchange_personal_data.orders_manager
    )
    assert isinstance(tbtp_group.active_order_swap_strategy, personal_data.StopFirstActiveOrderSwapStrategy)
    stop_loss = created_order(personal_data.StopLossOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss.update(
        price=decimal.Decimal(8),
        quantity=decimal.Decimal("0.6"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=tbtp_group,
        is_active=True,
        active_trigger=personal_data.create_order_price_trigger(stop_loss, decimal.Decimal(8), False),
        exchange_order_id="e_stop_loss_1"
    )
    stop_loss.trailing_profile = personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(price, price, True)
        for price in (20, 30)
    ])
    sell_limit_1 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_1.update(
        price=decimal.Decimal(20),
        quantity=decimal.Decimal("0.4"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=tbtp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_1, decimal.Decimal(20), True),
        exchange_order_id="e_sell_limit_1"
    )
    sell_limit_2 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_2.update(
        price=decimal.Decimal(30),
        quantity=decimal.Decimal("0.2"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=tbtp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_2, decimal.Decimal(30), True),
        exchange_order_id="e_sell_limit_2"
    )
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
    assert portfolio["BTC"].available == decimal.Decimal(10)
    locked_amount = constants.ZERO
    for order in [stop_loss, sell_limit_1, sell_limit_2]:
        await order.initialize()
        await order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order)
        if order in [stop_loss]:
            # stop sell limit quantity is not removed from available but stop is
            locked_amount += order.origin_quantity
        assert portfolio["BTC"].available == decimal.Decimal(10) - locked_amount
    assert stop_loss.is_active is True
    assert sorted(
        order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False), key=lambda x: x.origin_price
    ) ==  sorted(
        [sell_limit_1, sell_limit_2], key=lambda x: x.origin_price
    )
    assert len(tbtp_group.get_group_open_orders()) == 3
    assert all(order in tbtp_group.get_group_open_orders()
               for order in order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders())

    async def _edit_order(*args, quantity=None, price=None, **kwargs):
        edited_stop_loss = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                                        trader_instance, side=enums.TradeOrderSide.SELL)
        price = price or decimal.Decimal(10)
        edited_stop_loss.update(
            price=price,
            quantity=quantity,
            symbol=DEFAULT_ORDER_SYMBOL,
            order_type=enums.TraderOrderType.STOP_LOSS,
            group=tbtp_group,
            is_active=True,
            active_trigger=personal_data.create_order_price_trigger(edited_stop_loss, price, False),
            exchange_order_id="edited_stop_loss"
        )
        order_dict = edited_stop_loss.to_dict()
        order_dict.pop(enums.ExchangeConstantsOrderColumns.ID.value)
        return order_dict

    with mock.patch.object(exchange_manager.exchange, "cancel_order", mock.AsyncMock(return_value=enums.OrderStatus.CANCELED)) as cancel_order_mock, \
            mock.patch.object(exchange_manager.exchange, "edit_order",
                              mock.AsyncMock(side_effect=_edit_order)) as edit_order_mock:
        # STEP 1: filling sell_limit_1
        # sell_limit_1 become active: stop_loss is reduced and remains active
        step_1_maybe_partially_inactive_orders, step_1_reverse = await tbtp_group.adapt_before_order_becoming_active(sell_limit_1)
        sell_limit_1.is_active = True    # simulate now active order
        assert step_1_maybe_partially_inactive_orders == [stop_loss]
        cancel_order_mock.assert_not_called()
        edit_order_mock.assert_called_once()
        assert stop_loss.is_active is True
        for order in [stop_loss, sell_limit_1, sell_limit_2]:
            assert not order.is_cleared()
        all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
        assert len(all_orders) == 3
        assert len(tbtp_group.get_group_open_orders()) == 3
        assert all(order in tbtp_group.get_group_open_orders() for order in all_orders)
        assert stop_loss in all_orders
        assert all(order in all_orders for order in [sell_limit_1, sell_limit_2])
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert len(inactive_orders) == 1
        assert inactive_orders == [sell_limit_2]
        # stop_loss is reduced
        assert stop_loss.origin_quantity == decimal.Decimal("0.2")  # 0.6 - 0.4
        # stop_loss price is still the same (sell order is not filled)
        assert stop_loss.origin_price == decimal.Decimal("8")
        cancel_order_mock.reset_mock()
        edit_order_mock.reset_mock()

        # fill sell_limit_1: stop_loss price trails, no order is cancelled
        await sell_limit_1.on_fill(force_fill=True)
        assert sell_limit_1.is_closed()
        assert not stop_loss.is_closed()
        assert sell_limit_1.is_cleared()
        assert not stop_loss.is_cleared()
        cancel_order_mock.assert_not_called()
        # trailing
        edit_order_mock.assert_called_once()
        assert edit_order_mock.mock_calls[0].args[0] == stop_loss.exchange_order_id
        assert edit_order_mock.mock_calls[0].kwargs["quantity"] == decimal.Decimal("0.2")
        assert edit_order_mock.mock_calls[0].kwargs["price"] == decimal.Decimal("20")
        assert stop_loss.origin_price == decimal.Decimal("20")  # updated price
        assert stop_loss.origin_quantity == decimal.Decimal("0.2")  # same quantity

        all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
        assert len(all_orders) == 2
        assert all(order in all_orders for order in [stop_loss, sell_limit_2])
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert inactive_orders == [sell_limit_2]
        cancel_order_mock.reset_mock()
        edit_order_mock.reset_mock()

        # STEP 2: filling sell_limit_2
        # sell_limit_2 become active: stop loss gets canceled
        step_2_maybe_partially_inactive_orders, step_1_reverse = await tbtp_group.adapt_before_order_becoming_active(sell_limit_2)
        sell_limit_2.is_active = True    # simulate now active order
        assert step_2_maybe_partially_inactive_orders == [stop_loss]
        cancel_order_mock.assert_called_once()
        assert cancel_order_mock.mock_calls[0].args[0] == stop_loss.exchange_order_id
        edit_order_mock.assert_not_called()
        assert sell_limit_2.is_active is True
        assert stop_loss.is_active is False
        for order in [stop_loss, sell_limit_2]:
            assert not order.is_cleared()
        all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
        assert len(all_orders) == 2
        assert len(tbtp_group.get_group_open_orders()) == 2
        assert all(order in tbtp_group.get_group_open_orders() for order in all_orders)
        assert all(order in all_orders for order in [stop_loss, sell_limit_2])
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert inactive_orders == [stop_loss]
        # stop_loss is not reduced
        assert stop_loss.origin_quantity == decimal.Decimal("0.2")
        # stop_loss price is still the same
        assert stop_loss.origin_price == decimal.Decimal("20")
        cancel_order_mock.reset_mock()
        edit_order_mock.reset_mock()

        # fill sell_limit_2: stop_loss is canceled
        await sell_limit_2.on_fill(force_fill=True)
        assert sell_limit_2.is_closed()
        assert stop_loss.is_closed()
        assert sell_limit_2.is_cleared()
        assert stop_loss.is_cleared()
        all_orders = exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
        assert all_orders == []
        cancel_order_mock.assert_not_called()
        edit_order_mock.assert_not_called()
        # did not trail (got canceled instead)
        assert stop_loss.origin_price == decimal.Decimal("20")  # same price
        assert stop_loss.origin_quantity == decimal.Decimal("0.2")  # same quantity

@pytest.mark.asyncio
async def test_trailing_balanced_adapt_before_order_becoming_active_2_simultaneous_orders(simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    trader_instance.allow_artificial_orders = False
    trader_instance.enable_inactive_orders = True
    trader_instance.simulate = False
    exchange_manager.is_backtesting = True  # force update_order_status
    tbtp_group = personal_data.TrailingOnFilledTPBalancedOrderGroup(
        "name", exchange_manager.exchange_personal_data.orders_manager
    )
    assert isinstance(tbtp_group.active_order_swap_strategy, personal_data.StopFirstActiveOrderSwapStrategy)
    stop_loss_1 = created_order(personal_data.StopLossOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss_1.update(
        price=decimal.Decimal(8),
        quantity=decimal.Decimal("0.8"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=tbtp_group,
        is_active=True,
        active_trigger=personal_data.create_order_price_trigger(stop_loss_1, decimal.Decimal(8), False),
        exchange_order_id="e_stop_loss_1"

    )
    stop_loss_2 = created_order(personal_data.StopLossOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss_2.update(
        price=decimal.Decimal(10),
        quantity=decimal.Decimal("0.2"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=tbtp_group,
        is_active=True,
        active_trigger=personal_data.create_order_price_trigger(stop_loss_2, decimal.Decimal(10), False),
        exchange_order_id="e_stop_loss_2"
    )
    updated_stop_loss_2 = [stop_loss_2]
    sell_limit_1 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_1.update(
        price=decimal.Decimal(20),
        quantity=decimal.Decimal("0.4"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=tbtp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_1, decimal.Decimal(20), True),
        exchange_order_id="e_sell_limit_1"
    )
    sell_limit_2 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_2.update(
        price=decimal.Decimal(30),
        quantity=decimal.Decimal("0.2"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=tbtp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_2, decimal.Decimal(30), True),
        exchange_order_id="e_sell_limit_2"
    )
    sell_limit_3 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_3.update(
        price=decimal.Decimal(40),
        quantity=decimal.Decimal("0.4"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=tbtp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_3, decimal.Decimal(40), True),
        exchange_order_id="e_sell_limit_3"
    )
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
    assert portfolio["BTC"].available == decimal.Decimal(10)
    locked_amount = constants.ZERO
    for order in [stop_loss_1, stop_loss_2, sell_limit_1, sell_limit_2, sell_limit_3]:
        await order.initialize()
        await order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order)
        if order in [stop_loss_1, stop_loss_2]:
            # stop sell limit quantity is not removed from available but stop is
            locked_amount += order.origin_quantity
        assert portfolio["BTC"].available == decimal.Decimal(10) - locked_amount
    assert stop_loss_1.is_active is True
    assert stop_loss_2.is_active is True
    assert sorted(
        order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False), key=lambda x: x.origin_price
    ) ==  sorted(
        [sell_limit_1, sell_limit_2, sell_limit_3], key=lambda x: x.origin_price
    )
    assert len(tbtp_group.get_group_open_orders()) == 5
    assert all(order in tbtp_group.get_group_open_orders()
               for order in order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders())

    async def _edit_order(*args, quantity=None, price=None, **kwargs):
        edited_stop_loss = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                                        trader_instance, side=enums.TradeOrderSide.SELL)
        edited_stop_loss.update(
            price=price or decimal.Decimal(10),
            quantity=quantity,
            symbol=DEFAULT_ORDER_SYMBOL,
            order_type=enums.TraderOrderType.STOP_LOSS,
            group=tbtp_group,
            is_active=True,
            active_trigger=personal_data.create_order_price_trigger(edited_stop_loss, decimal.Decimal(10), False),
            exchange_order_id="edited_stop_loss_1"
        )
        order_dict = edited_stop_loss.to_dict()
        order_dict.pop(enums.ExchangeConstantsOrderColumns.ID.value)
        return order_dict

    with mock.patch.object(exchange_manager.exchange, "cancel_order", mock.AsyncMock(return_value=enums.OrderStatus.CANCELED)) as cancel_order_mock, \
            mock.patch.object(exchange_manager.exchange, "edit_order",
                              mock.AsyncMock(side_effect=_edit_order)) as edit_order_mock:
        async def step1_task():
            # STEP 1
            # sell_limit_1 become active: stop_loss_2 is now inactive, stop_loss_1 is reduced and remains active
            step_1_maybe_partially_inactive_orders, step_1_reverse = await tbtp_group.adapt_before_order_becoming_active(sell_limit_1)
            sell_limit_1.is_active = True    # simulate now active order
            assert step_1_maybe_partially_inactive_orders == [stop_loss_2, stop_loss_1]
            cancel_order_mock.assert_called_once()
            edit_order_mock.assert_called_once()
            assert stop_loss_1.is_active is True
            assert stop_loss_2.is_active is False
            for order in [stop_loss_1, stop_loss_2, sell_limit_1, sell_limit_2, sell_limit_3]:
                assert not order.is_cleared()
            # stop_loss_2 is now inactive as well, sell limit could be updated to come active (done in strategy)
            all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
            assert len(all_orders) == 5
            assert len(tbtp_group.get_group_open_orders()) == 5
            assert all(order in tbtp_group.get_group_open_orders() for order in all_orders)
            assert stop_loss_1 in all_orders and stop_loss_2 in all_orders
            assert all(order in all_orders for order in [sell_limit_1, sell_limit_2, sell_limit_3])
            inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
            assert len(inactive_orders) == 3
            assert stop_loss_2 in inactive_orders and all(order in inactive_orders for order in [sell_limit_2, sell_limit_3])
            # stop_loss_1 is reduced
            assert stop_loss_1.origin_quantity == decimal.Decimal("0.6")  # 0.8 - 0.2
            assert stop_loss_1.origin_price == decimal.Decimal(8)  # same price
            cancel_order_mock.reset_mock()
            edit_order_mock.reset_mock()

            # reverse step 1
            exchange_created_order = personal_data.StopLossOrder(trader_instance)
            exchange_created_order.update(
                order_type=enums.TraderOrderType.STOP_LOSS,
                symbol=DEFAULT_ORDER_SYMBOL,
                quantity=stop_loss_2.origin_quantity,
                price=stop_loss_2.origin_price,
                status=enums.OrderStatus.OPEN,
                exchange_order_id="recreated_stop_loss_2",
                fee={
                    enums.FeePropertyColumns.COST.value: 1,
                    enums.FeePropertyColumns.CURRENCY.value: "USDT",
                }
            )
            with mock.patch.object(order_factory, "create_order_instance_from_raw",
                              mock.Mock(return_value=exchange_created_order)) as create_order_instance_from_raw_mock, \
                 mock.patch.object(exchange_manager.exchange, "create_order",
                                   mock.AsyncMock(return_value=exchange_created_order)) as create_order_mock:
                await step_1_reverse(sell_limit_1, step_1_maybe_partially_inactive_orders)
                # stop loss 1 gets increased, stop loss 2 gets active again (re-created) and sell_limit_1 is back to inactive
                cancel_order_mock.assert_called_once()
                assert cancel_order_mock.mock_calls[0].args[0] == sell_limit_1.exchange_order_id
                edit_order_mock.assert_called_once()
                assert edit_order_mock.mock_calls[0].args[0] == stop_loss_1.exchange_order_id
                create_order_mock.assert_called_once()
                assert create_order_mock.mock_calls[0].kwargs["order_type"] == enums.TraderOrderType.STOP_LOSS
                create_order_instance_from_raw_mock.assert_called_once()
                assert stop_loss_2.is_cleared()
                assert stop_loss_1.origin_quantity == decimal.Decimal("0.8")
                assert stop_loss_1.origin_price == decimal.Decimal(8)  # same price
                assert exchange_created_order.origin_quantity == stop_loss_2.origin_quantity
                assert exchange_created_order.is_active is True
                updated_stop_loss_2[0] = exchange_created_order
                assert sell_limit_1.is_active is False
                all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
                assert len(all_orders) == 5
                assert all(order in all_orders for order in [stop_loss_1, updated_stop_loss_2[0]])
                assert len(tbtp_group.get_group_open_orders()) == 5
                assert all(order in tbtp_group.get_group_open_orders() for order in all_orders)
                inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
                assert len(inactive_orders) == 3
                assert all(order in inactive_orders for order in [sell_limit_1, sell_limit_2, sell_limit_3])
                cancel_order_mock.reset_mock()
                edit_order_mock.reset_mock()
                create_order_mock.reset_mock()
                create_order_instance_from_raw_mock.reset_mock()

        async def step2_task():
            # STEP 2 => because of the group order locks, should be executed after step 1
            # sell_limit_2 become active: updated_stop_loss_2[0] (created in step 1) is cancelled, stop_loss_1 has no change
            step_2_maybe_partially_inactive_orders, step_2_reverse = await tbtp_group.adapt_before_order_becoming_active(sell_limit_2)
            sell_limit_2.is_active = True    # simulate now active order
            assert step_2_maybe_partially_inactive_orders == [updated_stop_loss_2[0]]
            cancel_order_mock.assert_called_once()
            edit_order_mock.assert_not_called()
            assert stop_loss_1.is_active is True
            assert updated_stop_loss_2[0].is_active is False
            assert sell_limit_1.is_active is False
            assert sell_limit_2.is_active is True
            assert sell_limit_3.is_active is False
            for order in [stop_loss_1, updated_stop_loss_2[0], sell_limit_1, sell_limit_2, sell_limit_3]:
                assert not order.is_cleared()
            # updated_stop_loss_2[0] is now inactive as well, sell limit could be updated to come active (done in strategy)
            all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
            assert len(all_orders) == 5
            assert len(tbtp_group.get_group_open_orders()) == 5
            assert all(order in tbtp_group.get_group_open_orders() for order in all_orders)
            assert stop_loss_1 in all_orders and updated_stop_loss_2[0] in all_orders
            assert all(o in all_orders for o in [sell_limit_1, sell_limit_2, sell_limit_3])
            inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
            assert len(inactive_orders) == 3
            assert all(o in inactive_orders for o in [updated_stop_loss_2[0], sell_limit_1, sell_limit_3])
            # stop_loss_1 is NOT reduced
            assert stop_loss_1.origin_quantity == decimal.Decimal("0.8")
            cancel_order_mock.reset_mock()
            edit_order_mock.reset_mock()

            # reverse step 2
            exchange_created_order = personal_data.StopLossOrder(trader_instance)
            exchange_created_order.update(
                order_type=enums.TraderOrderType.STOP_LOSS,
                symbol=DEFAULT_ORDER_SYMBOL,
                quantity=stop_loss_2.origin_quantity,
                price=stop_loss_2.origin_price,
                status=enums.OrderStatus.OPEN,
                exchange_order_id="re-recreated_stop_loss_2",
                fee={
                    enums.FeePropertyColumns.COST.value: 1,
                    enums.FeePropertyColumns.CURRENCY.value: "USDT",
                }
            )
            with mock.patch.object(order_factory, "create_order_instance_from_raw",
                              mock.Mock(return_value=exchange_created_order)) as create_order_instance_from_raw_mock, \
                 mock.patch.object(exchange_manager.exchange, "create_order",
                                   mock.AsyncMock(return_value=exchange_created_order)) as create_order_mock:
                await step_2_reverse(sell_limit_2, step_2_maybe_partially_inactive_orders)
                # stop loss 1 does not change, stop loss 2 gets active again (re-created) and sell_limit_2 is back to inactive
                cancel_order_mock.assert_called_once()
                assert cancel_order_mock.mock_calls[0].args[0] == sell_limit_2.exchange_order_id
                edit_order_mock.assert_not_called()
                create_order_mock.assert_called_once()
                assert create_order_mock.mock_calls[0].kwargs["order_type"] == enums.TraderOrderType.STOP_LOSS
                create_order_instance_from_raw_mock.assert_called_once()
                assert updated_stop_loss_2[0].is_cleared()
                assert stop_loss_1.origin_quantity == decimal.Decimal("0.8")
                assert stop_loss_1.origin_price == decimal.Decimal(8)  # same price
                assert exchange_created_order.origin_quantity == updated_stop_loss_2[0].origin_quantity
                assert sell_limit_1.is_active is False
                assert exchange_created_order.is_active is True

                re_updated_stop_loss_2 = exchange_created_order
                assert sell_limit_1.is_active is False
                all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
                assert len(all_orders) == 5
                assert all(order in all_orders for order in [stop_loss_1, re_updated_stop_loss_2])
                assert len(tbtp_group.get_group_open_orders()) == 5
                assert all(order in tbtp_group.get_group_open_orders() for order in all_orders)

                inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
                assert len(inactive_orders) == 3
                assert all(order in inactive_orders for order in [sell_limit_1, sell_limit_2, sell_limit_3])
                cancel_order_mock.reset_mock()
                edit_order_mock.reset_mock()
                create_order_mock.reset_mock()
                create_order_instance_from_raw_mock.reset_mock()

        await asyncio.gather(
            step1_task(),
            step2_task(),
        )

def test_TrailingSideBalance_get_order_update(trailing_side_balance, order_f):
    order = order_f
    # no trailing profile
    assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
        personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
        personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_QUANTITY: decimal.Decimal(0),
        personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE: decimal.Decimal(0),
    }

    # incompatible trailing profile
    order.trailing_profile = mock.Mock()
    assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
        personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
        personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
        personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_QUANTITY: decimal.Decimal(0),
        personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE: decimal.Decimal(0),
    }

    order.trailing_profile = personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(price, price, True)
        for price in (10000, 12000, 13000)
    ])
    # filled price does not trigger trailing: trailing trigger price is not reached
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(9000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(8900), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_QUANTITY: decimal.Decimal(0),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE: decimal.Decimal(0),
        }
        get_potentially_outdated_price_mock.assert_called_once()

    # filled price triggers trailing: trailing trigger price is reached by closed order even though current price is bellow
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(10000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(8900), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: decimal.Decimal(10000),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_QUANTITY: decimal.Decimal(0),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE: decimal.Decimal(0),
        }
        get_potentially_outdated_price_mock.assert_called_once()

    # filled price does not trailing: trailing trigger price is reached by current price even though closed order is bellow
    # BUT trailing threshold has already been reached (trailed already)
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(1000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(10000), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: None,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_QUANTITY: decimal.Decimal(0),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE: decimal.Decimal(0),
        }
        get_potentially_outdated_price_mock.assert_called_once()

    # filled price triggers trailing: trailing trigger price is reached by current price even though closed order is bellow
    closed_order = personal_data.Order(order.trader)
    closed_order.origin_price = decimal.Decimal(1000)
    trailing_side_balance.closed_order = closed_order
    with mock.patch.object(
        order_util, "get_potentially_outdated_price", mock.Mock(return_value=(decimal.Decimal(12000), True))
    ) as get_potentially_outdated_price_mock:
        assert trailing_side_balance.get_order_update(order, decimal.Decimal(12)) == {
            personal_data.TrailingOnFilledTPBalancedOrderGroup.ORDER: order,
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_QUANTITY: decimal.Decimal(12),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE: decimal.Decimal(12000),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_QUANTITY: decimal.Decimal(0),
            personal_data.TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE: decimal.Decimal(0),
        }
        get_potentially_outdated_price_mock.assert_called_once()



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_commons.tree as commons_tree
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.html_util as html_util

import octobot_trading.errors as errors
import octobot_trading.personal_data.trades.channel as trades_channel
import octobot_trading.constants as constants
import octobot_trading.util as util


class TradesUpdater(trades_channel.TradesProducer):
    """
    The Trades Update fetch the exchange trades and send it to the Trade Channel
    """

    """
    The updater related channel name
    """
    CHANNEL_NAME = constants.TRADES_CHANNEL

    """
    Trades history request limit
    """
    MAX_OLD_TRADES_TO_FETCH = 100
    TRADES_LIMIT = 10

    """
    The default trade history update refresh time in seconds
    """
    TRADES_REFRESH_TIME = 5 * commons_constants.MINUTE_TO_SECONDS

    DEPENDENCIES_TIMEOUT = 5 * commons_constants.MINUTE_TO_SECONDS

    def __init__(self, channel):
        super().__init__(channel)

        self._is_initialized_event_set = False

    async def init_trade_history(self):
        try:
            await self.fetch_and_push()
            self._set_all_initialized()
            await asyncio.sleep(self.TRADES_REFRESH_TIME)
        except errors.NotSupported:
            self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
            await self.pause()
        except Exception as e:
            self.logger.error(f"Fail to initialize trade history : {html_util.get_html_summary_if_relevant(e)}")

    async def fetch_and_push(self):
        self.logger.debug(
            f"Updating {self.channel.exchange_manager.exchange_config.traded_symbol_pairs} trades history"
        )
        for symbol in self._get_pairs_to_update():
            if trades := await self.channel.exchange_manager.exchange.get_my_recent_trades(
                symbol=symbol,
                limit=self.MAX_OLD_TRADES_TO_FETCH
            ):
                await self.push(trades)

    def _set_all_initialized(self):
        for symbol in self._get_pairs_to_update():
            if not self._is_initialized_event_set:
                self._set_initialized_event(symbol)
        self._is_initialized_event_set = True

    def _set_initialized_event(self, symbol):
        # set init in updater as it's the only place we know if we fetched trades or not regardless of trades existence
        commons_tree.EventProvider.instance().trigger_event(
            self.channel.exchange_manager.bot_id, commons_tree.get_exchange_path(
                self.channel.exchange_manager.exchange_name,
                commons_enums.InitializationEventExchangeTopics.TRADES.value,
                symbol=symbol
            )
        )

    async def start(self):
        if util.is_trade_history_loading_enabled(self.channel.exchange_manager.config):
            await self.wait_for_dependencies(
                [
                    commons_tree.get_exchange_path(
                        self.channel.exchange_manager.exchange_name,
                        commons_enums.InitializationEventExchangeTopics.CONTRACTS.value
                    ),
                ],
                self.DEPENDENCIES_TIMEOUT
            )
            await self.init_trade_history()

        if self.channel.exchange_manager.exchange_config.get_forced_updater_channel_specs(self.CHANNEL_NAME):
            await self._run_update_loop()

    async def _run_update_loop(self):
        while not self.should_stop and not self.channel.is_paused:
            try:
                await self.fetch_and_push()
            except Exception as e:
                self.logger.error(f"Fail to update trades : {html_util.get_html_summary_if_relevant(e)}")

            await asyncio.sleep(self.TRADES_REFRESH_TIME)

    def _get_pairs_to_update(self):
        return self.channel.exchange_manager.exchange_config.traded_symbol_pairs + self.channel.exchange_manager.exchange_config.additional_traded_pairs

    async def resume(self) -> None:
        """
        Resume updater process
        """
        await super().resume()
        if not self.is_running:
            await self.run()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channels
import octobot_backtesting.api as backtesting_api


async def stop_and_pause(producer) -> None:
    """
    Stop and pause the provided producer
    :param producer: the producer to stop and pause
    """
    await super(type(producer), producer).stop()
    try:
        await producer.pause()
    except KeyError:
        pass
    producer.time_consumer = None


async def pause_time_consumer(producer) -> None:
    """
    Unregister the provided producer's time consumer
    :param producer: the producer to pause
    """
    if producer.time_consumer is not None:
        await get_time_channel(producer).remove_consumer(producer.time_consumer)


async def resume_time_consumer(producer, producer_time_callback) -> None:
    """
    Register the provided producer's time consumer
    :param producer: the producer to resume
    :param producer_time_callback: the producer TIME_CHANNEL callback
    """
    if producer.time_consumer is None and not producer.channel.is_paused:
        producer.time_consumer = await get_time_channel(producer).new_consumer(producer_time_callback)


def get_time_channel(producer):
    return channels.get_chan(
        backtesting_api.get_backtesting_time_channel_name(producer.channel.exchange_manager.exchange.backtesting)
    )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.symbols as symbols_util
import octobot_trading.enums as trading_enums
import octobot_trading.constants as constants
import octobot_trading.personal_data as personal_data


_LOSING_ORDER_TYPES = [
    trading_enums.TradeOrderType.STOP_LOSS,
    trading_enums.TradeOrderType.STOP_LOSS_LIMIT,
]


def compute_win_rate(exchange_manager):
    lost_trades_count = constants.ZERO
    won_trades_count = constants.ZERO
    entries = constants.ZERO
    if exchange_manager.is_future:
        for trade in exchange_manager.exchange_personal_data.trades_manager.trades.values():
            if trade.status is trading_enums.OrderStatus.FILLED:
                if trade.reduce_only:
                    if trade.exchange_trade_type in _LOSING_ORDER_TYPES:
                        lost_trades_count += constants.ONE
                    else:
                        won_trades_count += constants.ONE
                else:
                    entries += constants.ONE
        total_exits = lost_trades_count + won_trades_count
        if total_exits > entries:
            # multiple take profits and SL not handled yet
            win_rate = constants.ZERO
        else:
            win_rate = won_trades_count / total_exits if total_exits else constants.ZERO
        if win_rate != constants.ZERO:
            return win_rate
        else:  # try fallback to transactions for trailing stops and multiple exits
            lost_trades_count = constants.ZERO
            won_trades_count = constants.ZERO
            for transaction in exchange_manager.exchange_personal_data.transactions_manager.transactions.values():
                if isinstance(transaction, personal_data.RealisedPnlTransaction) and transaction.is_closed_pnl():
                    if transaction.realised_pnl > constants.ZERO:
                        won_trades_count += constants.ONE
                    else:
                        lost_trades_count += constants.ONE
    else:
        for trade in exchange_manager.exchange_personal_data.trades_manager.trades.values():
            if trade.status is trading_enums.OrderStatus.FILLED and trade.side is trading_enums.TradeOrderSide.SELL:
                if trade.exchange_trade_type in _LOSING_ORDER_TYPES:
                    lost_trades_count += constants.ONE
                else:
                    won_trades_count += constants.ONE
    total_counted_trades = won_trades_count + lost_trades_count
    if total_counted_trades > constants.ZERO:
        return won_trades_count / total_counted_trades
    return constants.ZERO


def aggregate_trades_by_exchange_order_id(trades: list) -> dict:
    aggregated_trades_by_exchange_order_id = {}
    for trade in trades:
        if trade.exchange_order_id not in aggregated_trades_by_exchange_order_id:
            # initialise as a new trade that will contain the aggregation
            aggregated_trades_by_exchange_order_id[trade.exchange_order_id] = personal_data.create_trade_from_dict(
                trade.trader, trade.to_dict()
            )
        else:
            # include other trades from the same order id into the aggregated trade
            to_update = aggregated_trades_by_exchange_order_id[trade.exchange_order_id]
            try:
                to_update.executed_price = (
                    to_update.executed_quantity * to_update.executed_price + trade.executed_quantity * trade.executed_price
                ) / (to_update.executed_quantity + trade.executed_quantity)
            except (decimal.DivisionByZero, decimal.InvalidOperation):
                to_update.executed_price = constants.ZERO
            to_update.executed_quantity += trade.executed_quantity
            to_update.total_cost += trade.total_cost
            to_update.fee[trading_enums.FeePropertyColumns.COST.value] \
                += trade.fee[trading_enums.FeePropertyColumns.COST.value]
            to_update.executed_time = max(to_update.executed_time, trade.executed_time)
    return aggregated_trades_by_exchange_order_id


def get_real_or_estimated_trade_fee(trade: "personal_data.Trade") -> tuple[dict, bool]:
    order_type = trading_enums.TraderOrderType.BUY_LIMIT if trade.side is trading_enums.TradeOrderSide.BUY else trading_enums.TraderOrderType.SELL_LIMIT
    trading_order_fee_currency = None
    trading_order_fee_rate = None
    is_estimated_trading_fee = False
    if trade.fee and all(
        key.value in trade.fee for key in [
            trading_enums.FeePropertyColumns.CURRENCY, trading_enums.FeePropertyColumns.RATE
        ]
    ):
        trading_order_fee_currency = trade.fee[trading_enums.FeePropertyColumns.CURRENCY.value]
        trading_order_fee_rate = trade.fee[trading_enums.FeePropertyColumns.RATE.value]
    else:
        # try to get fees from somewhere else
        try:
            # check if order has fees
            order = trade.exchange_manager.exchange_personal_data.orders_manager.get_order(
                None, exchange_order_id=trade.exchange_order_id
            )
            if not order.fee or not all(
                key.value in order.fee for key in [
                    trading_enums.FeePropertyColumns.CURRENCY, trading_enums.FeePropertyColumns.RATE
                ]
            ):
                raise KeyError("no fee")
            trading_order_fee_currency = order.fee[trading_enums.FeePropertyColumns.CURRENCY.value]
            trading_order_fee_rate = order.fee[trading_enums.FeePropertyColumns.RATE.value]
        except KeyError:
            # try with other trades from the same order
            trades = trade.exchange_manager.exchange_personal_data.trades_manager.get_trades(
                None, exchange_order_id=trade.exchange_order_id
            )
            for other_trade in trades:
                if not other_trade.fee or not all(
                    key.value in other_trade.fee for key in [
                        trading_enums.FeePropertyColumns.CURRENCY, trading_enums.FeePropertyColumns.RATE
                    ]
                ):
                    continue
                trading_order_fee_currency = other_trade.fee[trading_enums.FeePropertyColumns.CURRENCY.value]
                trading_order_fee_rate = other_trade.fee[trading_enums.FeePropertyColumns.RATE.value]

    if trading_order_fee_currency and trading_order_fee_rate:
        base, quote = symbols_util.parse_symbol(trade.symbol).base_and_quote()
        if trading_order_fee_currency == base:
            trading_order_fee_cost = trading_order_fee_rate * trade.executed_quantity
        elif trading_order_fee_currency == quote:
            trading_order_fee_cost = trading_order_fee_rate * trade.executed_quantity * trade.executed_price
        else:
            # fee currency is not the base or quote of the symbol
            trading_order_fee_cost = constants.ZERO
        trading_fee = {
            trading_enums.FeePropertyColumns.CURRENCY.value: trading_order_fee_currency,
            trading_enums.FeePropertyColumns.COST.value: trading_order_fee_cost,
        }
    else:
        # estimate fees
        trading_fee = trade.exchange_manager.exchange.get_trade_fee(
            trade.symbol, order_type, trade.executed_quantity, trade.executed_price, trade.taker_or_maker
        )
        is_estimated_trading_fee = True
    return trading_fee, is_estimated_trading_fee



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.constants as constants


async def create_order_state(
    order, is_from_exchange_data=False, enable_associated_orders_creation=True, 
    is_already_counted_in_available_funds=False, ignore_states=None
):
    if ignore_states is None:
        ignore_states = []

    if order.status is enums.OrderStatus.PENDING_CREATION \
       and enums.States.PENDING_CREATION not in ignore_states:
        await order.on_pending_creation(
            enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )
    elif order.status is enums.OrderStatus.OPEN and enums.States.OPEN not in ignore_states:
        await order.on_open(
            force_open=False, is_from_exchange_data=is_from_exchange_data,
            enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )
    elif order.status in constants.FILL_ORDER_STATUS_SCOPE \
            and enums.OrderStates.FILLED not in ignore_states \
            and enums.States.CLOSED not in ignore_states:
        await order.on_fill(
            force_fill=False, is_from_exchange_data=is_from_exchange_data,
            enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )
    elif order.status in constants.CANCEL_ORDER_STATUS_SCOPE and enums.OrderStates.CANCELED not in ignore_states:
        await order.on_cancel(
            force_cancel=False, is_from_exchange_data=is_from_exchange_data,
            enable_associated_orders_creation=enable_associated_orders_creation,
            is_already_counted_in_available_funds=is_already_counted_in_available_funds
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.orders.cancel_policies import cancel_policy_factory
from octobot_trading.personal_data.orders.cancel_policies.cancel_policy_factory import (
    create_cancel_policy,
)

from octobot_trading.personal_data.orders.cancel_policies import order_cancel_policy
from octobot_trading.personal_data.orders.cancel_policies.order_cancel_policy import (
    OrderCancelPolicy,
)

from octobot_trading.personal_data.orders.cancel_policies import expiration_time_order_cancel_policy
from octobot_trading.personal_data.orders.cancel_policies.expiration_time_order_cancel_policy import (
    ExpirationTimeOrderCancelPolicy,
)

from octobot_trading.personal_data.orders.cancel_policies import chained_order_filling_price_order_cancel_policy
from octobot_trading.personal_data.orders.cancel_policies.chained_order_filling_price_order_cancel_policy import (
    ChainedOrderFillingPriceOrderCancelPolicy,
)

__all__ = [
    "create_cancel_policy",
    "OrderCancelPolicy",
    "ExpirationTimeOrderCancelPolicy",
    "ChainedOrderFillingPriceOrderCancelPolicy",
]
